import AutoRPE.UtilsRPE.RegexPattern as RegexPattern
import AutoRPE.UtilsRPE.SourceManager as SourceManager
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import AutoRPE.UtilsRPE.Error as Error
import re


def randStr(N):
    """Generates a random string with specified length"""
    import string
    import random
    chars = string.ascii_uppercase + string.digits
    return ''.join(random.choice(chars) for _ in range(N))


def mask_literal(string, dict_of_substitutions=None):
    """ Search for open and close of different apex and mask what's in between"""
    if dict_of_substitutions is None:
        dict_of_substitutions = {}
    # Before checking for apex, check for apex with escapes
    escaped_string = re.finditer(r'\bb\'.*\'', string)
    for e_s in escaped_string:
        string = _mask(string, e_s.start(),  e_s.start() + len(e_s.group()), dict_of_substitutions)

    apex = [i for i in ["\"", "\\'", "\'"] if string.count(i)]
    # Need to check for the fist that appears to avoid masking things already masked '(A,".nc")'
    apex.sort(key=lambda x: string.index(x))
    for i in apex:
        apex = re.finditer(i, string)
        apex = [a.start() for a in apex]
        for i_open, i_close in zip(apex[0::2], apex[1::2]):
            string = _mask(string, i_open, i_close + 1, dict_of_substitutions)
    return string


def _mask(string, start, end, dict_of_sbs, placeholder=False):
    """ Generate hash and substitutes them in the specified place"""
    if not placeholder:
        # Generate a unique hash
        dict_key = _key_gen(end-start, dict_of_sbs)

        # Add the entry in the dictionary with hash as key and literal as value
        dict_of_sbs[dict_key] = string[start: end]
        # Substitute the hash into the original string
        string = string[:start] + dict_key + string[end:]
    else:
        # Here we just substitute with a safe string, the string will never be unmasked
        # Generate a unique hash
        dict_key = "A" * (end - start)

        # Add the entry in the dictionary with hash as key and literal as value
        dict_of_sbs[dict_key] = string[start: end]
        # Substitute the hash into the original string
        string = string[:start] + dict_key + string[end:]
    return string


def _key_gen(len_string, dict_of_sbs):
    """The hash can't be confused with name of function or var since they always contain a @"""
    # Generate an hash as long as the literal -1 to leave space for the "@"
    dict_key = randStr(len_string - 1)
    # Insert "@" as second char so we can still use the \bkey_dict\b regex
    dict_key = dict_key[:1] + '@' + dict_key[1:]

    # When key is short (3 char) it is possible that already exists, recreate until a new key is generated
    if dict_key in dict_of_sbs:
        return _key_gen(len_string, dict_of_sbs)
    else:
        return dict_key


def mask_keyword(line, key, dictionary_of_substitutions):
    """Check if a variables is called following a keyword with the same name"""
    import re
    key = RegexPattern.escape_string(key)
    match = re.finditer(r'(%s)\s*=\s*%s' % (key, key), line, re.I)
    match = [[m.start(1), m.end(1)] for m in match]
    for start, end in match:
        # Substitute the match with the unique placeholder
        line = _mask(line,  start, end, dictionary_of_substitutions)
    return line


def remove_literals(string, dict_of_literal_sub=None):
    if dict_of_literal_sub is None:
        dict_of_literal_sub = {}
    string = mask_literal(string, dict_of_literal_sub)
    # Substitute literals with space, maintaining length
    for key, val in zip(dict_of_literal_sub.keys(), dict_of_literal_sub.values()):
        string = re.sub(RegexPattern.escape_string(key), "\'" + " "*(len(val)-2) + "\'", string)
    return string


def mask_string_for_check_operation(string, dictionary_of_replacements=None):
    import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
    # Remove exponent in the form of d**N: it will mess up with normal operators. It will not be substituted back
    string = remove_pow(string)
    string = BasicFunctions.remove_unary_operator(string)
    # Dict to store all entities that will need a mask
    if dictionary_of_replacements is None:
        dictionary_of_replacements = {}
    # Mask the exponential form of numbers/variables (may contain operations in the exponent)
    string = mask_exponential(string, dictionary_of_replacements)

    # Mask all arrays and structures (may contain operation in indexing)
    string = mask_arrays_and_structures(string, dictionary_of_replacements=dictionary_of_replacements)

    # Mask all functions call (may contain operation in arguments)
    string = mask_functions(string, dictionary_of_replacements)

    # Now is safe to remove brackets
    string = string.replace("(", "")
    string = string.replace(")", "")
    return string


def mask_arrays_and_structures(string, only_structures=False, dictionary_of_replacements=None):
    """ Replace all array and structures-like patter, creating a dictionary with the substitutions"""
    import AutoRPE.UtilsRPE.NatureDeterminer as NatureDeterminer

    placeholder = False
    # In case we do not want to store the replacement: create a tmp dict
    if dictionary_of_replacements is None:
        dictionary_of_replacements = {}
        # In this case the masking can be done in a simpler way
        placeholder = True

    # Check if the entire string is an array
    if NatureDeterminer.is_array(string) and not only_structures:
        dict_key = randStr(len(string))
        dictionary_of_replacements[dict_key] = string
        return dict_key

    array_replacement = {}
    # Checks for all array pattern, and substitute with randomly generated string of the same length
    start = True
    while start:
        start = False
        array_match = re.finditer(RegexPattern.array, string)
        for a_m in array_match:
            start = True
            string = _mask(string, a_m.start(), a_m.end(), array_replacement)

        # Substitute structures with string
        struct_match = re.finditer(RegexPattern.structure, string)
        for s_m in struct_match:
            start = True
            string = _mask(string, s_m.start(), s_m.end(), dictionary_of_replacements, placeholder)

    if only_structures:
        # Keep structure masked
        string = unmask_from_dict(string, array_replacement)
    else:
        # Save array dictionary for further use
        dictionary_of_replacements.update(array_replacement)
    return string


def remove_pow(string):
    import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
    exponentials_var = re.finditer(RegexPattern.number_exponentiation, string)
    # When the var is elevate to simple power, remove power
    start_end = [[a.start(), a.end()] for a in exponentials_var]
    for match_start, match_end in start_end[::-1]:
        string = string[:match_start] + string[match_end:]
    # Exp elevated to some operation in brackets, remove all in brackets
    exponentials_var = re.finditer(r"\*\*\s*\(", string)
    start = [a.start() for a in exponentials_var]
    for match_start in start[::-1]:
        match_end = match_start + len(BasicFunctions.close_brackets(string[match_start:]))
        string = string[:match_start] + string[match_end:]
    return string


def mask_functions(string, dictionary_of_replacements=None):
    """ Once array and structures are masked, this will mask only function"""
    import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions

    if dictionary_of_replacements is None:
        dictionary_of_replacements = {}

    # Transform the (eventual) callable iterator into a list to iterate backward
    functions = [f for f in re.finditer(RegexPattern.call_to_function, string)]

    for f in functions[::-1]:
        # Find call
        function_call = BasicFunctions.close_brackets(string[f.start(1):])
        # Compute exact end of the function call
        f_end = f.start(1) + len(function_call)
        # Generate a placeholder with specified length
        string = _mask(string, f.start(1), f_end, dictionary_of_replacements)
    return string


def mask_exponential(string, dictionary_of_replacements=None):
    """ Replace all exponential creating a dictionary with the substitutions"""
    # If no dictionary is passed, create one
    if dictionary_of_replacements is None:
        dictionary_of_replacements = {}

    exponentials_DPE = re.finditer(RegexPattern.number_exp_DPE_notation, string, re.I)
    # Substitute match with string, group 1, which correspond to the actual exponential
    start_end = [[a.start(2), a.end(2)] for a in exponentials_DPE]
    for match_start, match_end in start_end[::-1]:
        # Generate a placeholder with specified length
        string = _mask(string, match_start, match_end, dictionary_of_replacements)
    return string


def mask_intrinsics(string):
    import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
    import AutoRPE.UtilsRPE.CallManager as CallManager
    import AutoRPE.UtilsRPE.NatureDeterminer as NatureDeterminer
    intrinsic = NatureDeterminer.has_intrinsics(string)

    while intrinsic:
        # Get the intrinsic call
        intrinsic_call = BasicFunctions.close_brackets(
            string[re.search(RegexPattern.name_occurrence % intrinsic, string, re.I).start():])
        called_arg = CallManager.find_call_arguments(intrinsic_call)
        arguments_to_remove = []
        if intrinsic.lower() in ['real', 'reshape']:
            # Remove loop index, or cast precision if present
            called_arg = called_arg[:1]
        for arg in called_arg:
            if arg.split("=")[0] in SourceManager.list_of_key:
                arguments_to_remove.append(arg)
        # Removed key specification
        for arg in arguments_to_remove:
            called_arg.remove(arg)
        # Rejoin pure arguments
        if len(called_arg) > 1:
            string = string.replace(intrinsic_call, "(" + ",".join(called_arg) + ")", re.I)
        else:
            string = string.replace(intrinsic_call, called_arg[0], re.I)
            string = BasicFunctions.remove_outer_brackets(string)
        # Check for further intrinsics
        intrinsic = NatureDeterminer.has_intrinsics(string)
    # Check if is a cast with precision and the precision was not specified via key
    called_arg = CallManager.find_call_arguments("(" + string + ")")
    if len(called_arg) == 2:
        if called_arg[1] in VariablePrecision.lookup_table:
            return [called_arg[0]]
        # It is the dimension ?
        if NatureDeterminer.is_a_number(called_arg[1]):
            return [called_arg[0]]
    if len(called_arg) > 1:
        raise Error.VariableNotFound("The string %s is not a variable" % string)
    return called_arg


def unmask_from_dict(string, dictionary_of_replacements):
    if not (isinstance(string, list)):
        str_list = [string]
    else:
        str_list = string
    for _index, _s in enumerate(str_list):
        keys = re.findall(r'\b\w*@\w*', _s)
        # Check that there is at least on element to substitute
        keys = [k for k in keys if k in dictionary_of_replacements.keys()]
        substitute = len(keys)
        while substitute:
            for k in keys:
                try:
                    val = dictionary_of_replacements[k]
                except KeyError:
                    # We hit a key from a different dictionary, avoid infinite loop
                    substitute = False
                    continue
                str_list[_index] = re.sub(RegexPattern.escape_string(k, strict=False), val, str_list[_index])
            # Update the matches
            keys = re.findall(r'\b\w*@\w*', str_list[_index])
            # Re-check if they are among dictionary keys
            substitute = any(k in keys for k in dictionary_of_replacements.keys())

    if not (isinstance(string, list)):
        return str_list[0]
    else:
        return str_list
